 /*
  * UAE - The Un*x Amiga Emulator
  *
  * GGI - Interface  
  *
  * Always remember: "GGI - The right thing to do[tm]"
  *
  * Copyright 1997 Christian Schmitt <schmittc@uni-freiburg.de>,
  *                Stefan Reinauer <stepan@www.schokola.de>
  */

#include "sysconfig.h"
#include "sysdeps.h"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <ggi/libggi.h>
#include <string.h>

#include "config.h"
#include "options.h"
#include "uae.h"
#include "memory.h"
#include "custom.h"
#include "readcpu.h"
#include "newcpu.h"
#include "keyboard.h"
#include "xwin.h"
#include "keybuf.h"
#include "gui.h"


static ggi_visual_t vis;
static ggi_graphtype type;
static int sx,sy,vx,vy;
ggi_color map[256],map2[256];
unsigned char pack[256];
static uae_u8 dither_buf[1000];
int buttonstate[3]={0,0,0};
int lastmx,lastmy,newmousecounters=0;
static int keystate[256];

xcolnr xcolors[4096];

int need_dither,bpp;
void *linear=NULL;
static int colors_allocated;

/* We don't need a brkhandler with GGI */
void setup_brkhandler(void)
{
}

static int get_color(int r, int g, int b, xcolnr *cnp)
{
   if(colors_allocated == 255) 	return -1;
   *cnp=colors_allocated;
   map[colors_allocated].r=doMask(r,16,0);
   map[colors_allocated].g=doMask(g,16,0);
   map[colors_allocated].b=doMask(b,16,0);
   ggiSetPaletteVec(vis,0,colors_allocated,map);
   colors_allocated++;
   return 1;
}

void init_colors(void)
{
    if(need_dither) {
	setup_dither(8,get_color);
    }else {
	int i;
	colors_allocated=0;
	gfxvidinfo.can_double=0;
	switch(bpp){
	  case  8: alloc_colors256(get_color);
		   for(i=0;i<4096;i++) xcolors[i]=xcolors[i]*0x01010101;
		   gfxvidinfo.can_double=1;
		   break;
	  case 16: alloc_colors64k(5,5,5,10,5,0);
		   for(i=0;i<4096;i++) xcolors[i]=xcolors[i]*0x00010001;
		   gfxvidinfo.can_double=1;
		   break;
	  case 32: alloc_colors64k(5,5,5,10,5,0); break;
	  default: break;
	}         
    }
}

int graphics_init(void) /* See how simple it is? :-> */
{
    int err,i;

    sx=vx=currprefs.gfx_width;
    sy=vy=currprefs.gfx_height;

    switch(currprefs.color_mode){
	case  0 : type=GT_8BIT;
		  need_dither=0;
		  break;
	case  1 :
	case  2 : type=GT_16BIT;
		  need_dither=0;
		  break;
	case  3 : type=GT_8BIT;
		  need_dither=1;
		  break;
	case  4 : type=GT_4BIT;
		  need_dither=1;
		  break;
        default : fprintf(stderr, "Bad color mode selected. Using default.\n"); 
        	  currprefs.color_mode = 0;
		  type=GT_8BIT;
		  break;
    }

    ggiInit();

    if (vis=ggiOpen("display-KGI:/dev/graph4")==NULL) {
        fprintf(stderr,"Can't open graphics console /dev/graph4\n");
        return 1;
    }

    ggiSetFocus(vis);
    if (err=ggiSetGraphMode(vis,sx,sy,vx,vy,type)) {
        fprintf(stderr,"Can't set graphics mode\n");
        return 2;
    }
    bpp=ggiGetBPP(vis);

    linear=ggiGetFB(vis);

    gfxvidinfo.maxlinetoscr = currprefs.gfx_width;
    gfxvidinfo.maxline = currprefs.gfx_height;
    gfxvidinfo.pixbytes = bpp/8;
    gfxvidinfo.rowbytes = gfxvidinfo.pixbytes * currprefs.gfx_width;
    if(linear!=NULL) gfxvidinfo.bufmem=linear; 
    else gfxvidinfo.bufmem = (char *)malloc(gfxvidinfo.rowbytes * currprefs.gfx_height);
    gfxvidinfo.maxblocklines=0;
    memset(gfxvidinfo.bufmem,0,gfxvidinfo.rowbytes * currprefs.gfx_height);
    for (i=0;i<256;i++) {
       map[i].r=i;
       map[i].g=i;
       map[i].b=i;
    }
    ggiPackColors(vis,pack,map,256);
    init_colors();
    buttonstate[0]=buttonstate[1]=buttonstate[2]=0;
    for(i=0;i<256;i++) keystate[i]=0;
    lastmx=lastmy=0;
    return 1;
}

void graphics_leave(void)
{
        ggiClose(vis);
        ggiExit();
}

void flush_line(int y)
{
   int target_y = y;
   int bit_unit=8;
   int i=0;
   char *addr=gfxvidinfo.bufmem+y*gfxvidinfo.rowbytes;
   if (target_y < 400 && target_y >= 0) {
      if (need_dither) {
           if(linear==NULL){
		DitherLine(dither_buf, (uae_u16 *)addr, 0, y, 
                  gfxvidinfo.maxlinetoscr, bit_unit);
           	addr = dither_buf;
      		ggiPutHLine(vis,0,target_y,gfxvidinfo.maxlinetoscr, addr);
	   } else {
		DitherLine(linear+gfxvidinfo.rowbytes*y, (uae_u16 *)addr,0,y,
			gfxvidinfo.maxlinetoscr,bit_unit);
	   }
      }
      if(linear==NULL) ggiPutHLine(vis,0,target_y,640, addr);
   }
}

void flush_block(int a, int b){};
void flush_screen(int a, int b){};

#define KEY_NOTHING 0x00 
#define AK_QUIT 0xf00

/* We should not need to have different versions for different keyboard */
static short decodekey[128]={
   KEY_NOTHING,	/* 0 */
   AK_ESC,
   AK_1, AK_2, AK_3, AK_4, AK_5, AK_6, AK_7, AK_8, AK_9, AK_0,
   AK_MINUS,	/* 12 */
   AK_EQUAL,
   AK_BS,
   AK_TAB,	/* 0x0f */
   AK_Q, AK_W, AK_E, AK_R, AK_T, 
   AK_Y,	/* 0x15  Attention: language specific !!!! */
   AK_U, AK_I, AK_O, AK_P, AK_LBRACKET,
   AK_RBRACKET,	/* 0x1b */
   AK_RET,
   AK_CTRL,
   AK_A, AK_S, AK_D, AK_F, AK_G, AK_H, AK_J, AK_K, AK_L,
   AK_SEMICOLON,	/* 0x27 */
   AK_QUOTE,	/* &auml; 0x28 */
   KEY_NOTHING,
   AK_LSH,	/* 0x2a */ 
   AK_LTGT,	/* < 0x2b Attention: language specific !!!! */
   AK_Z,	/* 0x2c Attention: language specific !!!! */
   AK_X, AK_C, AK_V, AK_B, AK_N, AK_M, AK_COMMA, AK_PERIOD, 
   AK_SLASH,	/* 0x35 */
   AK_RSH,	/* 0x36 */
   AK_NPMUL,
   AK_LALT,
   AK_SPC,
   AK_CAPSLOCK,	/* 0x3a */
   AK_F1, AK_F2, AK_F3, AK_F4, AK_F5, AK_F6, AK_F7, AK_F8, AK_F9,
   AK_F10,	/* 0x44 */
   KEY_NOTHING,	/* NumLock */
   KEY_NOTHING,	/* "Rollen" */
   AK_NP7,	/* 0x47 */
   AK_NP8,
   AK_NP9,
   AK_NPSUB,	/* 0x4a */
   AK_NP4,
   AK_NP5,
   AK_NP6,
   AK_NPADD,	/* 0x4e */
   AK_NP1,
   AK_NP2,
   AK_NP3,
   AK_NP0,	/* 0x52 */
   AK_NPDEL,
   KEY_NOTHING,	/* ??? */
   KEY_NOTHING,	/* ??? */
   KEY_NOTHING,	/* ??? */
   AK_BS,	/* 0x57 --> F11 */
   AK_QUIT,	/* 0x58 --> F12 */
   KEY_NOTHING,	/* ??? */
   KEY_NOTHING,	/* ??? */
   KEY_NOTHING,	/* ??? */
   KEY_NOTHING,	/* ??? */
   KEY_NOTHING,	/* ??? */
   KEY_NOTHING,	/* ??? */
   KEY_NOTHING,	/* ??? */
   AK_ENT,	/* 0x60 */
   AK_RCTRL,	/* 0x61 */
   AK_NPDIV,
   KEY_NOTHING,
   AK_RALT,	/* 0x64 */
   KEY_NOTHING,	/* ??? */
   AK_NP7,	/* Pos1 */
   AK_UP,	/* 0x67 */ 
   AK_LAMI,	/* 0x68  --> PgUp, */
   AK_LF,	/* 0x69 */
   AK_RT,	/* 0x6a */
   AK_NP1,	/* End */
   AK_DN,	/* 0x6c */
   AK_RAMI,	/* --> PgDown */
   AK_HELP,	/* 0x6e --> insert */
   AK_DEL,
   KEY_NOTHING,
   KEY_NOTHING,
   KEY_NOTHING,
   KEY_NOTHING,
   KEY_NOTHING,
   KEY_NOTHING,
   KEY_NOTHING,
   KEY_NOTHING,	/* 0x77 --> Pause */
   KEY_NOTHING,
   KEY_NOTHING,
   KEY_NOTHING,
   KEY_NOTHING,
   KEY_NOTHING,
   AK_LAMI,	/* 0x7d	--> WindOoze left  */
   AK_RAMI,	/*	--> WindOoze right */
   AK_CTRL	/* 0x7f	--> WindOoze Menu  */
};

void handle_events(void)
{
    struct timeval t={0,1};
    ggi_event ev;
    ggi_event_mask mask;
    int akey;

    while(ggiEventPoll(vis,emPtrButton,&t)){	
        uint32 state=0,
	       button=0;   
        ggiEventRead(vis,&ev,emPtrButton);
	state=ev.pbutton.state;
	button=ev.pbutton.button;
        switch(button) {
	    case 1 : buttonstate[2]=state&button;break;
            case 2 : buttonstate[1]=(state&button)>>1;break;
            case 4 : buttonstate[0]=(state&button)>>2;break;
	    default: break; 
	}
   } 

   while(ggiEventPoll(vis,emPtrRelative,&t)) {
        ggiEventRead(vis,&ev,emPtrRelative);
        lastmx+=ev.pmove.x;
	lastmy+=ev.pmove.y;
   }

   while(ggiEventPoll(vis,emKey,&t)) {
	ggiEventRead(vis,&ev,emKey);
        if(ev.key.sym==0) continue;
        if(ev.key.code>128) return;

        akey=decodekey[ev.key.code];

        if(akey==0xf00) uae_quit();

	if(keystate[AK_CTRL] && keystate[AK_LAMI] && keystate[AK_RAMI]) 
		m68k_reset();

        switch(ev.key.type) {
	    case 6 /* Rel */  : keystate[akey]=0;
			        record_key((akey << 1) | 1);
                                break;
            case 5 /* Press*/ : if(akey==KEY_NOTHING) break;
				if(!keystate[akey]) {
				   keystate[akey]=1;
                                   record_key(akey << 1);
				}
				break;
	    case 7 /*Repeat*/ : break;
	    default	      : break;
	}
    }
    return;
}

void write_log(const char *buf)
{
	fprintf(stderr,buf);
}

int debuggable(void)
{
	return 0;
}

int needmousehack(void)
{
	return 0;
}

/* This is not implemented yet as GGI is not that far */
void LED(int on)
{
}
