 /*
  * UAE - The Un*x Amiga Emulator
  *
  * Main program
  *
  * Copyright 1995 Ed Hanway
  * Copyright 1995, 1996, 1997 Bernd Schmidt
  */

#include "sysconfig.h"
#include "sysdeps.h"

#include <assert.h>
#include <ctype.h>

#include "config.h"
#include "options.h"
#include "gensound.h"
#include "sounddep/sound.h"
#include "events.h"
#include "uae.h"
#include "memory.h"
#include "custom.h"
#include "serial.h"
#include "readcpu.h"
#include "newcpu.h"
#include "disk.h"
#include "debug.h"
#include "xwin.h"
#include "joystick.h"
#include "target.h"
#include "keybuf.h"
#include "gui.h"
#include "zfile.h"
#include "autoconf.h"
#include "osemu.h"
#include "osdep/exectasks.h"
#include "compiler.h"

int version = 100*UAEMAJOR + 10*UAEMINOR + UAEURSAMINOR;
struct uae_prefs currprefs = {
/* framerate */ 1,
/* illegal_mem */ 0,
/* no_xhair */ 0,
/* use_serial */ 0,
#ifndef __DOS__
/* serial_demand */ 0,
/* parallel_demand */ 0,
#else
/* serial_demand */ 1,
/* parallel_demand */ 1,
#endif
/* automount_uaedev */ 1,

/* fake_joystick */ 2 + (0<<8),
/* KbdLang keyboard_lang */ KBD_LANG_US,
/* allow_save */ 0,
/* emul_accuracy */ 2,
/* test_drawing_speed */ 0,

/* produce_sound */ 0,
/* sound_bits */ DEFAULT_SOUND_BITS,
/* sound_freq */ DEFAULT_SOUND_FREQ,
/* sound_minbsiz */ DEFAULT_SOUND_MINB,
/* sound_maxbsiz */ DEFAULT_SOUND_MAXB,

/* gfx_width */ 800,
/* gfx_height */ 600,
/* gfx_lores */ 0,
/* gfx_linedbl */ 0,
/* gfx_correct_aspect */ 0,
/* gfx_xcenter */ 0,
/* gfx_ycenter */ 0,
/* color_mode */ 0,

/* use_low_bandwidth */ 0,
/* use_mitshm */ 0,

/* immediate_blits */ 0,
/* blits_32bit_enabled */ 0,

/* floppies */ { "df0.adf", "df1.adf", "df2.adf", "df3.adf" },
/* hardfile */ "hardfile",

/* copper_pos */ -1
};
struct uae_prefs changed_prefs;
unsigned int m68k_speed = 4;

int no_gui = 0, use_debugger = 0, use_gfxlib = 0;

long hardfile_size = 0;

int joystickpresent = 0;

char warning_buffer[256];

uae_u32 fastmem_size = 0x000000;
uae_u32 chipmem_size = 0x200000;
uae_u32 bogomem_size = 0x000000;
int address_space_24 = 1;
char romfile[256] = "kick.rom";
#ifndef __DOS__
char prtname[256] = "lpr ";
#else
char prtname[256] = "PRN:";
extern int DOS_init(void);
#endif

char optionsfile[256];

/* If you want to pipe printer output to a file, put something like
 * "cat >>printerfile.tmp" above.
 * The printer support was only tested with the driver "PostScript" on
 * Amiga side, using apsfilter for linux to print ps-data.
 *
 * Under DOS it ought to be -p LPT1: or -p PRN: but you'll need a
 * PostScript printer or ghostscript -=SR=-
 */

/* People must provide their own name for this */
#ifndef __DOS__
char sername[256] = "";
#else
char sername[256] = "AUX:";
#endif

/* Slightly stupid place for this... */
/* ncurses.c might use quite a few of those. */
char *colormodes[] = { "256 colors", "32768 colors", "65536 colors",
    "256 colors dithered", "16 colors dithered", "16 million colors",
    "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", ""
};

static void fix_options(void)
{
}

int quit_program = 0;

void uae_reset (void)
{
    if (quit_program != 1 && quit_program != -1)
	quit_program = -2;
}

void uae_quit (void)
{
    if (quit_program != -1)
	quit_program = -1;
}

/*
 * This function is dangerous, at least if you use different versions of UAE
 * from time to time (as I do). For example, running the SVGAlib version with
 * "-o" will discard the keyboard language information...
 */
void save_options(FILE *f)
{
    /* We only get here if allow_save is true, so... */
    fprintf(f, "-o\n");
    if (use_debugger)
	fprintf(f, "-D\n");
    fprintf(f, "-r %s\n", romfile);
    if (strcmp(currprefs.df[0], "")) fprintf(f, "-0 %s\n", currprefs.df[0]);
    if (strcmp(currprefs.df[1], "")) fprintf(f, "-1 %s\n", currprefs.df[1]);
    if (strcmp(currprefs.df[2], "")) fprintf(f, "-2 %s\n", currprefs.df[2]);
    if (strcmp(currprefs.df[3], "")) fprintf(f, "-3 %s\n", currprefs.df[3]);
#ifndef UNSUPPORTED_OPTION_p
    fprintf(f, "-p %s\n", prtname);
#endif
#ifndef UNSUPPORTED_OPTION_S
#ifndef __DOS__
    fprintf(f, "-S %d\n", currprefs.produce_sound);
#else
    {
    extern int sound_adjust, sound_stereo;
    fprintf(f, "-S %d", currprefs.produce_sound);
    if (sound_adjust) fprintf(f, "A");
    if (sound_stereo) fprintf(f, "S");
    fprintf(f, "\n");
    }
#endif
#endif
    if (currprefs.fake_joystick)
	fprintf(f, "-J %c%c\n", "01MABC"[currprefs.fake_joystick & 255],
		"01MABC"[(currprefs.fake_joystick >> 8) & 255]);
    fprintf(f, "-f %d\n", currprefs.framerate);
    fprintf(f, "-O %d:%d:", currprefs.gfx_width, currprefs.gfx_height);
    if (currprefs.gfx_lores)
	fprintf(f, "l");
    switch (currprefs.gfx_xcenter) {
     case 1: fprintf(f, "x"); break;
     case 2: fprintf(f, "X"); break;
     default: break;
    }
    switch (currprefs.gfx_ycenter) {
     case 1: fprintf(f, "y"); break;
     case 2: fprintf(f, "Y"); break;
     default: break;
    }
    if (currprefs.gfx_linedbl)
	fprintf(f, "d");
    if (currprefs.gfx_correct_aspect)
	fprintf(f, "c");
    fprintf(f, "\n-H %d\n", currprefs.color_mode);
    if (fastmem_size > 0)
	fprintf(f, "-F %d\n", fastmem_size / 0x100000);
    if (bogomem_size > 0)
	fprintf(f, "-s %d\n", bogomem_size / 0x40000);
    fprintf(f, "-c %d\n", chipmem_size / 0x80000);
    if (!address_space_24)
	fprintf(f, "-4\n");
    if (!currprefs.automount_uaedev)
	fprintf(f, "-a\n");
    if (currprefs.use_low_bandwidth)
	fprintf(f, "-L\n");
    if (currprefs.use_mitshm)
	fprintf(f, "-T\n");
#ifndef UNSUPPORTED_OPTION_B
    fprintf(f, "-B %d:%d\n", currprefs.sound_minbsiz, currprefs.sound_maxbsiz);
#endif
#ifndef UNSUPPORTED_OPTION_b
    fprintf(f, "-b %d\n", currprefs.sound_bits);
#endif
#ifndef UNSUPPORTED_OPTION_R
    fprintf(f, "-R %d\n", currprefs.sound_freq);
#endif
    fprintf(f, "-A %d\n", currprefs.emul_accuracy);
    if (currprefs.immediate_blits || currprefs.blits_32bit_enabled) {
	fprintf(f, "-n ");
	if (currprefs.immediate_blits)
	    fprintf(f, "i");
	if (currprefs.blits_32bit_enabled)
	    fprintf(f, "3");
	fprintf(f, "\n");
    }
    if (use_gfxlib)
	fprintf(f, "-g\n");
    if (currprefs.copper_pos != -1)
	fprintf(f, "-C %d\n", currprefs.copper_pos);
    fprintf(f, "-w %d\n", m68k_speed);
    /* We don't write "-t" - I can hardly imagine a user who wants that in his
     * config file. */
    write_filesys_config(f);
#ifndef UNSUPPORTED_OPTION_l
    fprintf(f, "-l %s\n", (currprefs.keyboard_lang == KBD_LANG_DE ? "de"
			   : currprefs.keyboard_lang == KBD_LANG_ES ? "es"
			   : currprefs.keyboard_lang == KBD_LANG_US ? "us"
			   : currprefs.keyboard_lang == KBD_LANG_SE ? "se"
			   : currprefs.keyboard_lang == KBD_LANG_FR ? "fr"
			   : currprefs.keyboard_lang == KBD_LANG_IT ? "it"
			   : "FOO"));
#endif
}

struct optiondef {
    const char *getopt_str;
    const char *helpstr;
};


static struct optiondef uaeopts[] = {
    { "h",	  "  -h           : Print help\n" },
#ifndef UNSUPPORTED_OPTION_l
    { "l:",	  "  -l lang      : Set keyboard language to lang, where lang is\n"
		  "                 DE, SE, US, ES, FR or IT\n" },
#endif
    { "v:",       "  -v VOL:dir   : mount a virtual directory called <dir> as AmigaDOS volume\n"
		  "                 VOL:, this option supports full amiga file names and\n"
		  "                 attributes\n" },
    { "V:",       "  -V VOL:dir   : like -v, but mount read-only\n" },
    { "m:",	  "  -m VOL:dir   : mount directory called <dir> as AmigaDOS volume VOL:\n" },
    { "M:",	  "  -M VOL:dir   : like -m, but mount read-only\n" },
    { "s:",	  "  -s n         : Emulate n*256 KB slow memory at 0xC00000\n" },
    { "c:",	  "  -c n         : Emulate n*512 KB chip memory at 0x000000\n" },
    { "F:",	  "  -F n         : Emulate n MB fast memory at 0x200000\n" },
#if CPU_LEVEL > 1
    { "4",        "  -4           : Use a 32 bit address space instead of 24 bit.\n" },
#endif
    { "a",	  "  -a           : Add no expansion devices (disables fastmem and\n"
		  "                 harddisk support\n" },
    { "J:",	  "  -J xy        : Specify how to emulate joystick port 0 (x) and 1 (y)\n"
		  "                 Use 0 for joystick 0, 1 for joystick 1, M for mouse,\n"
		  "                 a/b/c for various keyboard replacements\n" },
    { "f:",	  "  -f n         : Set the frame rate to 1/n (draw only every nth frame)\n" },
#ifndef UNSUPPORTED_OPTION_D
    { "D",	  "  -D           : Start up the built-in debugger\n" },
#endif
    { "i",	  "  -i           : Print illegal memory accesses\n" },
    { "t",	  "  -t           : Test drawing speed (makes the emulator very slow)\n" },
#ifndef UNSUPPORTED_OPTION_o
    { "o",	  "  -o           : Allow options to be saved\n" },
#endif
#ifndef UNSUPPORTED_OPTION_G
    { "G",	  "  -G           : Disable user interface\n" },
#endif
    { "A:",	  "  -A n         : Set emulator accuracy to n (0, 1 or 2)\n" },
#ifdef USE_EXECLIB
    { "g",	  "  -g           : Turn on gfx-lib replacement (EXPERIMENTAL).\n" },
#endif
    { "C:",	  "  -C hpos      : Adjust copper emulation (see README for explanation)\n" },
    { "n:",	  "  -n parms     : Set blitter parameters: 'i' enables immediate blits,\n"
		  "                 '3' enables 32 bit blits (may crash RISC machines)\n" },
    { "0:1:2:3:", "  -[0123] file : Use file instead of df[0123].adf as disk image\n" },
    { "r:",	  "  -r file      : Use file as ROM image instead of kick.rom\n" },
#ifndef UNSUPPORTED_OPTION_S
#ifndef __DOS__
    { "S:",	  "  -S n         : Sound emulation accuracy (n = 0, 1, 2 or 3)\n"
		  "                 For sound emulation, n = 2 is recommended\n" },
#else
    { "S:",	  "  -S soundparm : Sound emulation parameters (see below)\n" },
#endif
#endif
#ifndef UNSUPPORTED_OPTION_b
    { "b:",	  "  -b n         : Use n bits for sound output (8 or 16)\n" },
#endif
#ifndef UNSUPPORTED_OPTION_R
    { "R:",	  "  -R n         : Use n Hz to output sound. Common values are\n"
		  "                 22050 Hz or 44100 Hz\n" },
#endif
#ifndef UNSUPPORTED_OPTION_B
    { "B:",	  "  -B n:m       : Specifies sound buffer size. n is the minimum value that\n"
		  "                 should be used, m is the maximum. On some machines, only\n"
		  "                 the maximum is used. Use small values on fast machines.\n" },
#endif
#ifndef UNSUPPORTED_OPTION_p
    { "p:",	  "  -p prt       : Use <prt> for printer output (default " DEFPRTNAME ").\n" },
#endif
#ifndef UNSUPPORTED_OPTION_I
    { "I:",	  "  -I device    : Use <device> for serial output (e.g. " DEFSERNAME ").\n" },
#ifndef __DOS__
    { "d:",	  "  -s [sp]      : Open Serial/Parallel port on demand only.\n" },
#endif
    { "w:",       "  -w n         : Set M68K_SPEED, large values for games, small for cpu apps,\n"
		  "                 default is 4.\n" },
#endif
#ifdef TARGET_SPECIAL_OPTIONS
    TARGET_SPECIAL_OPTIONS
#endif
    { "O:",	  "  -O modespec  : Define graphics mode (see below)\n" },
    { "H:",	  "  -H mode      : Set the color mode (see below)\n" }
};

void usage(void)
{
    int i;

    printf("UAE - The Un*x Amiga emulator\n");
    printf("Summary of command-line options (please read the README for more details):\n");
    for (i = 0; i < sizeof uaeopts / sizeof *uaeopts; i++)
	printf("%s", uaeopts[i].helpstr);
    
    printf("\n");
#ifdef __DOS__
    printf("%s",
	   "The format for the soundparm of \"-S\" is as follows:\n"
	   "  One number (0 to 3), to specify the sound emulation accuracy,\n"
	   "  for sound emulation, is recommended.\n"
	   "  Optionaly the folowing characters can be added for diferent features:\n"
	   "   A:    Adjust frequency, distorcing samples to reduce choppy sound (for\n"
	   "         slow machines).\n"
	   "   S:    Play in stereo sound.\n"
	   );
    printf("\n");
#endif
    printf("%s",
#ifndef COLOR_MODE_HELP_STRING
	   "Valid color modes: 0 (256 colors); 1 (32768 colors); 2 (65536 colors)\n"
	   "                   3 (256 colors, with dithering for better results)\n"
	   "                   4 (16 colors, dithered); 5 (16 million colors)\n"
#else
	   COLOR_MODE_HELP_STRING
#endif
	   );
    printf("%s",
	   "The format for the modespec parameter of \"-O\" is as follows:\n"
	   "  -O width:height:modifiers\n"
	   "  \"width\" and \"height\" specify the dimensions of the picture.\n"
	   "  \"modifiers\" is a string that contains zero or more of the following\n"
	   "  characters:\n"
	   "   l:    Treat the display as lores, drawing only every second pixel\n"
	   "   x, y: Center the screen horizontally or vertically.\n"
	   "   d:    Doubles the height of the display for better interlace emulation\n"
	   "   c:    Correct aspect ratio\n"
	   "UAE may choose to ignore the color mode setting and/or adjust the\n"
           "video mode setting to reasonable values.\n");
}

static void parse_gfx_specs (char *spec)
{
    char *x0 = my_strdup (spec);
    char *x1, *x2;
    
    x1 = strchr (x0, ':');
    if (x1 == 0)
	goto argh;
    x2 = strchr (x1+1, ':');
    if (x2 == 0)
	goto argh;
    *x1++ = 0; *x2++ = 0;
    
    currprefs.gfx_width = atoi (x0);
    currprefs.gfx_height = atoi (x1);
    currprefs.gfx_lores = strchr (x2, 'l') != 0;
    currprefs.gfx_xcenter = strchr (x2, 'x') != 0 ? 1 : strchr (x2, 'X') != 0 ? 2 : 0;
    currprefs.gfx_ycenter = strchr (x2, 'y') != 0 ? 1 : strchr (x2, 'Y') != 0 ? 2 : 0;
    currprefs.gfx_linedbl = strchr (x2, 'd') != 0;
    currprefs.gfx_correct_aspect = strchr (x2, 'c') != 0;
    
    free (x0);
    return;

    argh:
    fprintf (stderr, "Bad display mode specification.\n");
    fprintf (stderr, "The format to use is: \"width:height:modifiers\"\n");
    free (x0);
}

static void parse_sound_spec (char *spec)
{
    char *x0 = my_strdup (spec);
    char *x1 = strchr (x0, ':');
    if (x1 != NULL) {
	*x1++ = '\0'; currprefs.sound_maxbsiz = atoi (x1);
    } else
	currprefs.sound_maxbsiz = atoi (x0);
    currprefs.sound_minbsiz = atoi (x0);
    free (x0);
}

static int parse_joy_spec (char *spec)
{
    int v0 = 2, v1 = 0;
    if (strlen(spec) != 2)
	goto bad;

    switch (spec[0]) {
     case '0': v0 = 0; break;
     case '1': v0 = 1; break;
     case 'M': case 'm': v0 = 2; break;
     case 'A': case 'a': v0 = 3; break;
     case 'B': case 'b': v0 = 4; break;
     case 'C': case 'c': v0 = 5; break;
     default: goto bad;
    }

    switch (spec[1]) {
     case '0': v1 = 0; break;
     case '1': v1 = 1; break;
     case 'M': case 'm': v1 = 2; break;
     case 'A': case 'a': v1 = 3; break;
     case 'B': case 'b': v1 = 4; break;
     case 'C': case 'c': v1 = 5; break;
     default: goto bad;
    }
    if (v0 == v1)
	goto bad;
    /* Let's scare Pascal programmers */ 
    if (0)
bad:
    fprintf (stderr, "Bad joystick mode specification. Use -J xy, where x and y\n"
	     "can be 0 for joystick 0, 1 for joystick 1, M for mouse, and\n"
	     "a, b or c for different keyboard settings.\n");

    return v0 + (v1 << 8);
}

#ifndef DONT_PARSE_CMDLINE

void parse_cmdline(int argc, char **argv)
{
    int c, i;
    char getopt_str[256];
    strcpy (getopt_str, "");

    for (i = 0; i < sizeof uaeopts / sizeof *uaeopts; i++)
	strcat (getopt_str, uaeopts[i].getopt_str);

#ifdef __BEOS__
    if (argc < 2)
	return;
#endif
    /* Help! We're running out of letters! */
    while(((c = getopt(argc, argv, getopt_str)) != EOF))
    switch(c) {
     case 'h': usage();	exit(0);

     case '0': strncpy(currprefs.df[0], optarg, 255); currprefs.df[0][255] = 0; break;
     case '1': strncpy(currprefs.df[1], optarg, 255); currprefs.df[1][255] = 0; break;
     case '2': strncpy(currprefs.df[2], optarg, 255); currprefs.df[2][255] = 0; break;
     case '3': strncpy(currprefs.df[3], optarg, 255); currprefs.df[3][255] = 0; break;
     case 'r': strncpy(romfile, optarg, 255); romfile[255] = 0; break;
     case 'p': strncpy(prtname, optarg, 255); prtname[255] = 0; break;
     case 'I': strncpy(sername, optarg, 255); sername[255] = 0; currprefs.use_serial = 1; break;
     case 'm':
     case 'M':
     case 'v':
     case 'V':
	{
	    char buf[256];
	    char *s2, *ptr;
	    int readonly = isupper(c);
	    int vfs_module = (tolower(c) == 'v');
	    int vfs_mangle_size = 8;

	    strncpy(buf, optarg, 255); buf[255] = 0;
	    s2 = strchr(buf, ':');
	    if(s2) {
		*s2++ = '\0';
		if (vfs_module) {
		    ptr = strchr(s2, '*');
		    if (ptr) {
		       *ptr++ = '\0';
		       if ((*ptr) != 0)
			   vfs_mangle_size = atoi(ptr);
		       if ((vfs_mangle_size<1) ||  (vfs_mangle_size>8))
			   vfs_mangle_size = 8;
		    }
		}
#ifdef __DOS__
		{
		char *tmp;
		while ((tmp = strchr(s2, '\\')))
		    *tmp = '/';
		}
#endif
		add_filesys_unit(buf, s2, readonly, vfs_module, vfs_mangle_size);
	    } else {
		if (!vfs_module)
		    fprintf(stderr, "Usage: [-m | -M] VOLNAME:/mount_point\n");
		else
		    fprintf(stderr, "Usage: [-v | -V] VOLNAME:/mount_point*mangle_size\n");
	    }
	}
	break;

#ifndef __DOS__
     case 'S': currprefs.produce_sound = atoi (optarg); break;
#else
     case 'S':
	{
	    int i;
	    extern int sound_adjust, sound_stereo;

	    if (!optarg)
		break;
	    for (i=0; i<strlen(optarg); i++)
		switch(optarg[i]) {
		 case '0':
		 case '1':
		 case '2':
		 case '3': currprefs.produce_sound = optarg[i] - '0'; break;
		 case 'A':
		 case 'a': sound_adjust = 1; break;
		 case 'S':
		 case 's': sound_stereo = 1; break;
		}
	}
	break;
#endif
     case 'f': currprefs.framerate = atoi (optarg); break;
     case 'A': currprefs.emul_accuracy = atoi (optarg); break;
     case 'x': currprefs.no_xhair = 1; break;
     case 'i': currprefs.illegal_mem = 1; break;
     case 'J': currprefs.fake_joystick = parse_joy_spec (optarg); break;
     case 'a': currprefs.automount_uaedev = 0; break;
     case 'o': currprefs.allow_save = 1; break;
     case 't': currprefs.test_drawing_speed = 1; break;
     case 'L': currprefs.use_low_bandwidth = 1; break;
     case 'T': currprefs.use_mitshm = 1; break;

     case 'g': use_gfxlib = 1; break;
     case 'G': no_gui = 1; break;
     case 'D': use_debugger = 1; break;

     case 'n':
	if (strchr (optarg, '3') != 0)
	    currprefs.blits_32bit_enabled = 1;
	if (strchr (optarg, 'i') != 0)
	    currprefs.immediate_blits = 1;
	break;

     case 'C':
	currprefs.copper_pos = atoi(optarg);
	if (currprefs.copper_pos < -1 || currprefs.copper_pos > 224) {
	    fprintf(stderr, "Bad copper position specified (must be 0..224 or -1)\n");
	    currprefs.copper_pos = -1;
	}
	break;

     case 'F':
	fastmem_size = atoi(optarg) * 0x100000;
	if (fastmem_size != 0 && fastmem_size != 0x100000 && fastmem_size != 0x200000 
	    && fastmem_size != 0x400000 && fastmem_size != 0x800000) 
	{
	    fastmem_size = 0;
	    fprintf(stderr, "Unsupported fastmem size!\n");
	}
	break;

     case 's':
	bogomem_size = atoi(optarg) * 0x40000;
	if (bogomem_size != 0 && bogomem_size != 0x80000 && bogomem_size != 0x100000
	    /* Braino && bogomem_size != 0x180000 && bogomem_size != 0x1C0000*/)
	{
	    bogomem_size = 0;
	    fprintf(stderr, "Unsupported bogomem size!\n");
	}
	break;

     case 'c':
	chipmem_size = atoi(optarg) * 0x80000;
	if (chipmem_size != 0x80000 && chipmem_size != 0x100000
	    && chipmem_size != 0x200000 && chipmem_size != 0x400000
	    && chipmem_size != 0x800000)
	{
	    chipmem_size = 0x200000;
	    fprintf(stderr, "Unsupported chipmem size!\n");
	}

	break;

     case '4':
	address_space_24 = 0;
	break;

     case 'l':
	if (0 == strcasecmp(optarg, "de"))
	    currprefs.keyboard_lang = KBD_LANG_DE;
	else if (0 == strcasecmp(optarg, "us"))
	    currprefs.keyboard_lang = KBD_LANG_US;
	else if (0 == strcasecmp(optarg, "se"))
	    currprefs.keyboard_lang = KBD_LANG_SE;
	else if (0 == strcasecmp(optarg, "fr"))
	    currprefs.keyboard_lang = KBD_LANG_FR;
	else if (0 == strcasecmp(optarg, "it"))
	    currprefs.keyboard_lang = KBD_LANG_IT;
	else if (0 == strcasecmp(optarg, "es"))
	    currprefs.keyboard_lang = KBD_LANG_ES;
	break;

     case 'O': parse_gfx_specs (optarg); break;
     case 'd':
#ifndef __DOS__
	if (index(optarg,(int)'S')!=NULL || index(optarg,(int)'s')) {
		  fprintf (stdout, "  Serial on demand.\n");
		  currprefs.serial_demand=1;
	}
	if (index(optarg,(int)'P')!=NULL || index(optarg,(int)'p')) {
		  fprintf (stdout, "  Parallel on demand.\n");
		  currprefs.parallel_demand=1;
	}
#else
	fprintf (stderr, "Note: The -d option is obsolete, please use the new -O format.\n");
	c = atoi(optarg);
	switch (c) {
	 case 0:
	    parse_gfx_specs("320:200:lx");
	    break;
	 case 1:
	    parse_gfx_specs("320:240:lx");
	    break;
	 case 2:
	    parse_gfx_specs("320:400:lx");
	    break;
	 case 3:
	    parse_gfx_specs("640:300:x");
	    break;
	 default:
	    fprintf(stderr, "Bad video mode selected. Using default.\n");
	    /* fall through */
	 case 4:
	    parse_gfx_specs("800:600:");
	    break;
	}
#endif
	break;
     case 'H':
	currprefs.color_mode = atoi(optarg);
	if (currprefs.color_mode < 0) {
	    fprintf(stderr, "Bad color mode selected. Using default.\n");
	    currprefs.color_mode = 0;
	}
	break;

     case 'b': currprefs.sound_bits = atoi(optarg); break;
     case 'B': parse_sound_spec(optarg); break;
     case 'R': currprefs.sound_freq = atoi(optarg); break;
     case 'w': m68k_speed = atoi(optarg); break;
    }
}
#endif

static void parse_cmdline_and_init_file(int argc, char **argv)
{
    FILE *f;
    char *tmp;
    int t;
    char *home;
    char *buffer,*tmpbuf, *token;
    char smallbuf[256];
    int bufsiz, result;
    int n_args;
    char **new_argv;
    int new_argc;

    strcpy(optionsfile,"");

#ifdef OPTIONS_IN_HOME
    home = getenv("HOME");
    if (home != NULL && strlen(home) < 240)
    {
	strcpy(optionsfile, home);
	strcat(optionsfile, "/");
    }
#endif

    strcat(optionsfile, OPTIONSFILENAME);

    f = fopen(optionsfile, "rb");
#ifdef OPTIONS_IN_HOME
    /* sam: if not found in $HOME then look in current directory */
    if (f == NULL) {
	strcpy (optionsfile, OPTIONSFILENAME);
	f = fopen(optionsfile, "rb");
    }
#endif

    if (f == NULL) {
	parse_cmdline(argc, argv);
	return;
    }
    fseek(f, 0, SEEK_END);
    bufsiz = ftell(f);
    fseek(f, 0, SEEK_SET);

    buffer = (char *)malloc(bufsiz+1);
    buffer[bufsiz] = 0;
    if (fread (buffer, 1, bufsiz, f) < bufsiz) {
	fprintf (stderr, "Error reading configuration file\n");
	fclose (f);
	parse_cmdline (argc, argv);
	return;
    }
    fclose(f);

    /* Canonicalize the buffer */
    while ((tmp = strchr (buffer, '\r')))
	*tmp = ' ';
    for (;;) {
	tmp = strchr (buffer, '\n');
	if (tmp == 0)
	    break;
	if (tmp[1] == '#') {
	    char *tmp2 = strchr (tmp + 1, '\n');
	    if (tmp2 == 0)
		*tmp = '\0';
	    else
		memmove (tmp, tmp2, strlen (tmp2) + 1);
	} else {
	    *tmp = ' ';
	}
    }
    while ((tmp = strchr(buffer, '\t')))
	*tmp = ' ';

    tmp = buffer;
    while (*tmp == ' ')
	tmp++;
    t = strlen (tmp);
    memmove (buffer, tmp, t + 1);

    while (t > 0 && buffer[t-1] == ' ')
	buffer[--t] = '\0';
    for (;;) {
	tmp = strstr (buffer, "  ");
	if (tmp == 0)
	    break;
	memmove (tmp, tmp + 1, strlen (tmp));
    }

    tmpbuf = my_strdup (buffer);

    n_args = 0;
    if (strtok(tmpbuf, "\n ") != NULL) {
	do {
	    n_args++;
	} while (strtok(NULL, "\n ") != NULL);
    }
    free (tmpbuf);

    new_argv = (char **)malloc ((1 + n_args + argc) * sizeof (char **));
    new_argv[0] = argv[0];
    new_argc = 1;

    token = strtok(buffer, "\n ");
    while (token != NULL) {
	new_argv[new_argc] = my_strdup (token);
	new_argc++;
	token = strtok(NULL, "\n ");
    }
    for (n_args = 1; n_args < argc; n_args++)
	new_argv[new_argc++] = argv[n_args];
    new_argv[new_argc] = NULL;
    parse_cmdline(new_argc, new_argv);
}

/* Okay, this stuff looks strange, but it is here to encourage people who
 * port UAE to re-use as much of this code as possible. Functions that you
 * should be using are do_start_program() and do_leave_program(), as well
 * as real_main(). Some OSes don't call main() (which is braindamaged IMHO,
 * but unfortunately very common), so you need to call real_main() from
 * whatever entry point you have. You may want to write your own versions
 * of start_program() and leave_program() if you need to do anything special.
 * Add #ifdefs around these as appropriate.
 */

void do_start_program(void)
{
    /* Do a reset on startup. Whether this is elegant is debatable. */
    quit_program = 2;
    m68k_go(1);
}

void do_leave_program(void)
{
    graphics_leave();
    close_joystick();
    close_sound();
    dump_counts();
    serial_exit();
    zfile_exit();
    if (!no_gui)
	gui_exit();
}

#ifndef _WIN32
void start_program(void)
{
    do_start_program();
}
#endif

void leave_program(void)
{
    do_leave_program();
}

void real_main(int argc, char **argv)
{
    FILE *hf;

    if (!graphics_setup()) {
	exit(1);
    }

#ifdef __DOS__
    DOS_init();
#endif
    hf = fopen(currprefs.hf0, "rb");
    if (hf == NULL)
	hardfile_size = 0;
    else {
	fseek(hf, 0, SEEK_END);
	hardfile_size = ftell(hf);
	fclose(hf);
    }

    rtarea_init ();
    hardfile_install ();

    parse_cmdline_and_init_file(argc, argv);

    if (!setup_sound()) {
	fprintf(stderr, "Sound driver unavailable: Sound output disabled\n");
	currprefs.produce_sound = 0;
    }
    init_joystick();

    if (!no_gui) {
	int err = gui_init();
	if (err == -1) {
	    fprintf(stderr, "Failed to initialize the GUI\n");
	} else if (err == -2) {
	    exit(0);
	}
    }
    if (sound_available && !init_sound()) {
	fprintf(stderr, "Sound driver unavailable: Sound output disabled\n");
	currprefs.produce_sound = 0;
    }
	
    fix_options();
    changed_prefs = currprefs;
    
    /* Install resident module to get 8MB chipmem, if requested */
    rtarea_setup();
    
    keybuf_init(); /* Must come after init_joystick */
    
    expansion_init ();
    memory_init();

    filesys_install();
    gfxlib_install();
    emulib_install();
    trackdisk_install();

    custom_init(); /* Must come after memory_init */
    serial_init();
    DISK_init();
    init_m68k();
    compiler_init();
    gui_update();

    machdep_init();

    if (graphics_init ()) {
	setup_brkhandler ();
	filesys_start_threads ();
	if (use_debugger && debuggable ())
#ifndef __DOS__
	    activate_debugger();
#else
	    { extern void EnterDebug(void); EnterDebug(); }
#endif

	start_program ();
    }
    leave_program ();
}

#ifndef NO_MAIN_IN_MAIN_C
int main(int argc, char **argv)
{
    real_main(argc, argv);
    return 0;
}
#endif /* not __BEOS__ */
