 /*
  * UAE - The Un*x Amiga Emulator
  *
  * DosUAE Interface
  *
  * (c) 1996, 1997 Tim Gunn, Gustavo Goedert, Tomi Hakala
  *
  */

#include "sysconfig.h"
#include "sysdeps.h"

#include <conio.h>
#include <dpmi.h>

#define bstring other_bstring

#include "config.h"
#include "options.h"
#include "uae.h"
#include "autoconf.h"
#include "gui.h"
#include "memory.h"
#include "events.h"

#undef bstring

#include "video/dos-ui.h"
#include "video/dos-supp.h"
#include "video/vbe.h"
#define TB textbackground
#define TC textcolor

int gfxmenusort(const void *a, const void *b);
void save_options(FILE *f);

extern int menu_cur;
extern int sound_adjust, sound_stereo;

char textemp[100];

char adfdefaultdir[256];
char romdefaultdir[256];
char otherdefaultdir[256];

int quit_program;
int hdsize;
int kblang;

int origrows;

void drawscreen(void);

void HDOptions(void);
void DiskOptions(void);
void VideoOptions(void);
void MemoryOptions(void);
void SoundOptions(void);
void OtherOptions(void);

int selvol(void);
void aboutuae(void);
void tui_statusline(char*);
char *strcut(char *p,int num);

struct bstring *gfxmenu;
int *gfxitemstbl;
int numgfxitems = 0;

int gfxmenusort(const void *a, const void *b) {
   struct bstring *ga, *gb;

   ga = (struct bstring *) a;
   gb = (struct bstring *) b;
   if (ModeList[ga->val].ModeWidth < ModeList[gb->val].ModeWidth)
      return(-1);
   else if (ModeList[ga->val].ModeWidth > ModeList[gb->val].ModeWidth)
      return(1);
   else if (ModeList[ga->val].ModeHeight < ModeList[gb->val].ModeHeight)
      return(-1);
   else if (ModeList[ga->val].ModeHeight > ModeList[gb->val].ModeHeight)
      return(1);
   else if (ModeList[ga->val].ModeType < ModeList[gb->val].ModeType)
      return(-1);
   else if (ModeList[ga->val].ModeType > ModeList[gb->val].ModeType)
      return(1);
   return(0);
}

void gui_setup() {
    int i, j, hasres;
    struct bstring gfxmenuitem;

    getcwd(adfdefaultdir,256);
    getcwd(romdefaultdir,256);
    getcwd(otherdefaultdir,256);

    gfxmenu = (struct bstring *)calloc(NumberOfModes, sizeof(struct bstring));
    gfxitemstbl = (int *)calloc(NumberOfModes, sizeof(int));

    for (i = 0; i < NumberOfModes; i++) {
	hasres = 0;
	for (j = 0; j < numgfxitems; j++)
	    if ((ModeList[gfxmenu[j].val].ModeType == ModeList[i].ModeType) &&
		(ModeList[gfxmenu[j].val].ModeWidth == ModeList[i].ModeWidth) &&
		(ModeList[gfxmenu[j].val].ModeHeight == ModeList[i].ModeHeight))
		hasres = 1;
	if (!hasres) {
	    sprintf(gfxmenu[numgfxitems].data, "%dx%d ", ModeList[i].ModeWidth, ModeList[i].ModeHeight);
	    switch (ModeList[i].ModeType) {
		case T16:
		    strcat(gfxmenu[numgfxitems].data, "16 colors");
		    break;
		case T256:
		    strcat(gfxmenu[numgfxitems].data, "256 colors");
		    break;
		case T32K:
		    strcat(gfxmenu[numgfxitems].data, "32k colors");
		    break;
		case T64K:
		    strcat(gfxmenu[numgfxitems].data, "64k colors");
		    break;
		case T16M:
		    strcat(gfxmenu[numgfxitems].data, "16M colors");
		    break;
	    }
	    gfxmenu[numgfxitems].val = i;
	    numgfxitems++;
	}
    }
    qsort((void *)gfxmenu, numgfxitems, sizeof(struct bstring), gfxmenusort);
    for (i = 0; i < numgfxitems; i++) {
	gfxitemstbl[i] = gfxmenu[i].val;
	gfxmenu[i].val = -1;
    }
}

void gui_led(int led, int on)
{
}

int gui_update(void)
{
 return 0;
}


void gui_filename(int num, const char *name)
{
}

static void getline(char *p)
{
}

void gui_handle_events(void)
{
}

void tui_statusline(char *info)
{
    gotoxy(3,2);

    sprintf(textemp,"DosUAE v.%d - ",version);
    cprintf("%s",textemp);
    gotoxy(18,2);
    cprintf("                                                  ");
    gotoxy(18,2);
    cprintf("%s",info);
}

char *strcut(char *p,int num)
{
    char *tempbuf=(char *)calloc(num+1,sizeof(char));
    strncpy(tempbuf,p,num);
    return tempbuf;
}

void gui_exit()
{
    FILE *fp;
    char inChar;

    if (currprefs.allow_save) {
	FILE *f;
	f = fopen(optionsfile, "w");
	if (f == NULL) {
	    fprintf(stderr, "Error saving options file!\n");
	    return;
	}
	save_options(f);
	fclose(f);
    }
    tui_shutdown();
    tui_curson();

    _set_screen_lines(origrows);

    printf("have a nice day!\n");
}

int gui_init() {
    int i,sel=0,run=0;

    struct bstring *menu;
    menu = (struct bstring *)calloc(9 , sizeof(struct bstring));

    origrows = ScreenRows();

    tui_supp_init();
    gui_setup();
    quit_program = 0;

    i = open("hardfile", 1);
	hdsize = filelength(i)/1024;
	close(i);

    Update(1);
    tui_cursoff();
    menu_cur = 0;
    while(!run){
	strcpy(menu[0].data,"RUN UAE!");
	strcpy(menu[1].data,"Disk settings");
	strcpy(menu[2].data,"Video settings");
	strcpy(menu[3].data,"Memory settings");
	strcpy(menu[4].data,"Hard Disk settings");
	strcpy(menu[5].data,"Sound settings");
	strcpy(menu[6].data,"Other settings");
	strcpy(menu[7].data,"About UAE");
	strcpy(menu[8].data,"Exit");

	tui_statusline("Main Menu");

	sel = tui_menubrowse(menu,9);
	switch(sel) {
	    case 0:
		run=1;
	    break;
	    case 1:
		DiskOptions();
	    break;
	    case 2:
		VideoOptions();
	    break;
	    case 3:
		MemoryOptions();
	    break;
	    case 4:
		HDOptions();
	    break;
	    case 5:
		SoundOptions();
	    break;
	    case 6:
		OtherOptions();
	    break;
	    case 7:
		aboutuae();
	    break;
	    case 8:
		gui_exit();
		exit(0);
	    break;
	    case -1:
		gui_exit();
		exit(0);
	    break;
	}
	Update(0);
	}

    normvideo();
    clrscr();

    return 0;
}

void drawscreen(void)
{
    int i;
    tui_reset();
    tui_scrolloff(); tui_drawbox(1,1,80,50);
    tui_statusline("zed");
    gotoxy(3,3); cprintf("");
    for(i=0;i<17;i++) { gotoxy(28,i+4); cprintf(""); }
    for(i=0;i<11;i++) { gotoxy(28,i+24); cprintf(""); }
    gotoxy(3,19); cprintf("");
    gotoxy(28,19); cprintf("");
    gotoxy(3,21); cprintf("");
    gotoxy(28,21); cprintf("");
    gotoxy(28,3); cprintf("");
}

void Update(int type)
{
    int i, j;

    if(type==1 || type==2) drawscreen();
    if(type==0 || type==1)
    {
	tui_cleararea(31,4,79,34);
	gotoxy(31,4);
	if(currprefs.df[0][0]!=0) cprintf("DF0:- %s",strcut(currprefs.df[0],41));
	else cprintf("DF0:- none");
	gotoxy(31,5);
	if(currprefs.df[1][0]!=0) cprintf("DF1:- %s",strcut(currprefs.df[1],41));
	else cprintf("DF1:- none");
	gotoxy(31,6);
	if(currprefs.df[2][0]!=0) cprintf("DF2:- %s",strcut(currprefs.df[2],41));
	else cprintf("DF2:- none");
	gotoxy(31,7);
	if(currprefs.df[3][0]!=0) cprintf("DF3:- %s",strcut(currprefs.df[3],41));
	else cprintf("DF3:- none");

	gotoxy(31,9);
	cprintf("VIDEO: %dx%d", currprefs.gfx_width, currprefs.gfx_height);
	if(currprefs.gfx_lores==1)cprintf(" (lores)");
	cprintf(" %s ",colormodes[currprefs.color_mode]);

	gotoxy(33,10);
	switch (currprefs.gfx_xcenter) {
	    case 0: cprintf("                     "); break;
	    case 1: cprintf("X centered           "); break;
	    case 2: cprintf("X centered (clever)  "); break;
	}
	switch (currprefs.gfx_ycenter) {
	    case 0: cprintf("                   "); break;
	    case 1: cprintf("Y centered         "); break;
	    case 2: cprintf("Y centered (clever)"); break;
	}

	gotoxy(33,11);
	if(currprefs.gfx_correct_aspect==1)cprintf("aspect on  ");
	    else cprintf("           ");
	if(currprefs.gfx_linedbl==1)cprintf("double on  ");
	    else cprintf("           ");
	cprintf("drawing every ");
	switch(currprefs.framerate)
	{
	    case 1: break;
	    case 2: cprintf("2nd "); break;
	    case 3: cprintf("3rd "); break;
	    default: cprintf("%dth ",currprefs.framerate); break;
	}
	cprintf("frame.");

    gotoxy(31,13);
	if(currprefs.immediate_blits) cprintf("Immediate Blits On. ");
	else cprintf("Immediate Blits Off. ");
	if(currprefs.blits_32bit_enabled) cprintf("32 bit blits On.");
	else cprintf("32 bit blits Off.");

    gotoxy(31,15);
	if (currprefs.copper_pos == -1)
	    cprintf("Copper Cheat disabled");
	else
	    cprintf("Copper Cheat at: %d", currprefs.copper_pos);

    gotoxy(31,17);
	if(use_gfxlib==1)cprintf("GFXLIB replacement On");
	    else cprintf("GFXLIB replacement Off");

    gotoxy(31,19);
	cprintf("MEMORY: %dK chip, %dK fast, %dK bogo",chipmem_size/1024,fastmem_size/1024,bogomem_size/1024);

    gotoxy(31,21);
	cprintf("ROM IMAGE: %s",strcut(romfile,30));

    gotoxy(31,23);
	switch (currprefs.produce_sound) {
	    case 0: cprintf("SOUND: 0 (Off)                   "); break;
	    case 1: cprintf("SOUND: 1 (Off, but emulated)     "); break;
	    case 2: cprintf("SOUND: 2 (On)                    "); break;
	    case 3: cprintf("SOUND: 3 (On, emulated perfectly)"); break;
	}
    gotoxy(31,24);
	if (sound_stereo)
	    cprintf("  %d bits   %d Hz   Stereo     ", currprefs.sound_bits, currprefs.sound_freq);
	else
	    cprintf("  %d bits   %d Hz              ", currprefs.sound_bits, currprefs.sound_freq);


    gotoxy(31,25);
	cprintf("  Buffer: Min %d Max %d           ", currprefs.sound_minbsiz, currprefs.sound_maxbsiz);

    gotoxy(31,26);
	if (sound_adjust)
	    cprintf("  Adjust frequency dynamicly        ");
	else
	    cprintf("  Don't adjust frequency dynamicly  ");

    for(i=28;i<33;i++) { gotoxy(55,i); cprintf(""); }

    gotoxy(31,28);
	cprintf("JOYSTICK 0 CONFIG: %c", "01MABC"[currprefs.fake_joystick & 255]);
    gotoxy(57,28);
	cprintf("JOYSTICK 1 CONFIG: %c", "01MABC"[(currprefs.fake_joystick >> 8) & 255]);

    gotoxy(31,30);
	if(currprefs.use_serial){ cprintf("SERIAL: %s (active)  ",sername); }
	else { cprintf("SERIAL: %s (inactive)",sername); }
    gotoxy(57,30);
	cprintf("PARALLEL: %s",prtname);

    gotoxy(31,32);
	cprintf("ACCURACY: %d",currprefs.emul_accuracy);
    gotoxy(57,32);
	cprintf("68K SPEED: %d",m68k_speed);

    gotoxy(31,34);
	cprintf("HARDDISK: ");
	if(hdsize>16)
	{
	    cprintf("Hard disk file = %dKB is",hdsize);
	    if(currprefs.automount_uaedev==1)cprintf(" active    ");
		else cprintf(" not active");
	}
	else cprintf("No Hard disk file available");


    gotoxy(4,36); TC(SETHL); cprintf("UAE FILESYS VOLUMES"); TC(SETTC);
	for (i=0;i<13;i++) {
	    gotoxy(4,37+i);
	    cprintf("                                                                       ");
	}
	for (i=j=0;j<13;i++,j++) {
	    char buf[256];

	    gotoxy(4,37+j);
	    if (sprintf_filesys_unit(buf, i) == -1)
		break;
	    if (strncmp(buf, "Hardfile \"UAE", 13) != 0)
		cprintf(buf);
	    else
		j--;
	}
    gotoxy(4,26); TC(SETHL); cprintf("BUILD OPTIONS"); TC(SETTC);
    gotoxy(4,27);
    switch(CPU_LEVEL)
    {
	case 0: cprintf("CPU 68000 noFPU"); break;
	case 1: cprintf("CPU 68010 noFPU"); break;
	case 2: cprintf("CPU 68020 noFPU"); break;
	case 3: cprintf("CPU 68020/68881"); break;
    }
    gotoxy(4,28);
    #ifdef USE_COMPILER
    cprintf("68K COMPILER Version");
    #elif defined(NO_EXCEPTION_3)
    cprintf("FAST Version");
    #else
    cprintf("COMPATIBLE Version");
    #endif
    gotoxy(4,30); TC(SETHL); cprintf("PC SYSTEM"); TC(SETTC);
    gotoxy(4,31);cprintf("Physical Mem: %dMB",_go32_dpmi_remaining_physical_memory()/1048576);
    gotoxy(4,32);cprintf("Virtual Mem: %dMB",_go32_dpmi_remaining_virtual_memory()/1048576);
    gotoxy(4,33);cprintf("OS: %s",os_ver_str());
   }
}

void ChangeOps(void)
{
    int sel=0;
    char buf[256];

    struct bstring *menu;
    menu = (struct bstring *)calloc(2 , sizeof(struct bstring));

    drawscreen();

    menu[0].val=0; strcpy(menu[0].data,"Disk Options");
    strcpy(menu[1].data,"Return to UAE");

    menu_cur = 0;
    while(sel!=-1)
    {
	tui_statusline("Variable Options");
	Update(0);
	sel = tui_menubrowse(menu,2);

	switch(sel)
	{
	    case 0: DiskOptions(); break;
	    case 1: return; break;
	}
    }
}

void DiskOptions(void)
{
    int sel=0;
    char buf[256];

    struct bstring *menu;
    menu = (struct bstring *)calloc(9 , sizeof(struct bstring));

    menu[0].val=9; strcpy(menu[0].data,"Insert DF0");
    menu[1].val=-1; strcpy(menu[1].data,"Clear DF0");
    menu[2].val=9; strcpy(menu[2].data,"Insert DF1");
    menu[3].val=-1; strcpy(menu[3].data,"Clear DF1");
    menu[4].val=9; strcpy(menu[4].data,"Insert DF2");
    menu[5].val=-1; strcpy(menu[5].data,"Clear DF2");
    menu[6].val=9; strcpy(menu[6].data,"Insert DF3");
    menu[7].val=-1; strcpy(menu[7].data,"Clear DF3");
    strcpy(menu[8].data,"Main Menu");

    menu_cur = 0;
    while(sel!=-1)
    {
	tui_statusline("Disk Menu");
	Update(0);
	sel = tui_menubrowse(menu,9);

	switch(sel)
	{
	    case 0: strcpy(currprefs.df[0], tuifilereq(".adf",adfdefaultdir)); break;
	    case 1: strcpy(currprefs.df[0],""); break;
	    case 2: strcpy(currprefs.df[1], tuifilereq(".adf",adfdefaultdir)); break;
	    case 3: strcpy(currprefs.df[1],""); break;
	    case 4: strcpy(currprefs.df[2], tuifilereq(".adf",adfdefaultdir)); break;
	    case 5: strcpy(currprefs.df[2],""); break;
	    case 6: strcpy(currprefs.df[3], tuifilereq(".adf",adfdefaultdir)); break;
	    case 7: strcpy(currprefs.df[3],""); break;
	    case 8: sel=-1; break;
	}
    }
    menu_cur = 1;
}

void VideoOptions(void)
{

    int sel=0;

    struct bstring *menu;
    menu = (struct bstring *)calloc(13 , sizeof(struct bstring));

    strcpy(menu[0].data,"Video Mode");
    strcpy(menu[1].data,"Low resolution on/off");
    strcpy(menu[2].data,"Dithered on/off");
    strcpy(menu[3].data,"X center mode");
    strcpy(menu[4].data,"Y center mode");
    strcpy(menu[5].data,"Aspect correct on/off");
    strcpy(menu[6].data,"Double lines on/off");
    strcpy(menu[7].data,"Immediate Blits");
    strcpy(menu[8].data,"32 Bit Blits");
    strcpy(menu[9].data,"Frame Rate");
    strcpy(menu[10].data,"Copper Cheat");
    strcpy(menu[11].data,"GFXLIB on/off");
    strcpy(menu[12].data,"Main Menu");

    menu_cur = 0;
    while(sel!=-1)
    {
	tui_statusline("Video Menu");
	Update(0);
	sel = tui_menubrowse(menu,13);

	switch(sel)
	{
	    case 0:
		{
		  int vsel;

		  menu_cur = 0;
		  vsel = tui_menubrowse(gfxmenu, numgfxitems);
		  menu_cur = 0;
		  if (vsel != -1) {
		      currprefs.gfx_width = ModeList[gfxitemstbl[vsel]].ModeWidth;
		      currprefs.gfx_height = ModeList[gfxitemstbl[vsel]].ModeHeight;
		      switch (ModeList[gfxitemstbl[vsel]].ModeType) {
			  case T16:
			      currprefs.color_mode = 4;
			      break;
			  case T256:
			      currprefs.color_mode = 0;
			      break;
			  case T32K:
			      currprefs.color_mode = 1;
			      break;
			  case T64K:
			      currprefs.color_mode = 2;
			      break;
			  case T16M:
			      currprefs.color_mode = 5;
			      break;
		      }
		      currprefs.gfx_lores=currprefs.gfx_width<640;
		      currprefs.gfx_xcenter=2;
		      currprefs.gfx_ycenter=2;
		      currprefs.gfx_linedbl=0;
		      if (currprefs.gfx_height > 285)
			  currprefs.gfx_linedbl = 1;
		      currprefs.gfx_correct_aspect = 1;
		      if (currprefs.gfx_height > 570)
			  currprefs.gfx_correct_aspect = 0;
		  }
		}
	    break;
	    case 1:
		currprefs.gfx_lores=!currprefs.gfx_lores;
	    break;
	    case 2:
		if (currprefs.color_mode == 0)
		    currprefs.color_mode = 3;
		else if (currprefs.color_mode == 3)
		    currprefs.color_mode = 0;
	    break;
	    case 3:
		currprefs.gfx_xcenter++;
		if(currprefs.gfx_xcenter>2) { currprefs.gfx_xcenter=0; }
	    break;
	    case 4:
		currprefs.gfx_ycenter++;
		if(currprefs.gfx_ycenter>2) { currprefs.gfx_ycenter=0; }
	    break;
	    case 5:
		currprefs.gfx_correct_aspect=!currprefs.gfx_correct_aspect;
	    break;
	    case 6:
		currprefs.gfx_linedbl=!currprefs.gfx_linedbl;
	    break;
	    case 7:
		currprefs.immediate_blits=!currprefs.immediate_blits;
	    break;
	    case 8:
		currprefs.blits_32bit_enabled=!currprefs.blits_32bit_enabled;
	    break;
	    case 9:
		currprefs.framerate++;
		if(currprefs.framerate>9) { currprefs.framerate=1; }
	    break;
	    case 10:
	    {
		char copper_buf[22];

		*copper_buf = '\0';
		currprefs.copper_pos = -1;
		if(tuiscan("Enter Copper Cheat (0..224 or -1 to disable):",20,copper_buf))
		   currprefs.copper_pos = atoi(copper_buf);
		if (currprefs.copper_pos < -1 || currprefs.copper_pos > 224) {
		    currprefs.copper_pos = -1;
		}
	    }
	    break;
	    case 11:
		use_gfxlib=!use_gfxlib;
	    break;
	    case 12: sel=-1; break;
	}
    }
    menu_cur = 2;
}

void MemoryOptions() {

    int sel=0;

    struct bstring *menu;
    menu = (struct bstring *)calloc(5 , sizeof(struct bstring));

    strcpy(menu[0].data,"Fast Ram");
    strcpy(menu[1].data,"Chip Ram");
    strcpy(menu[2].data,"Bogo Ram");
    strcpy(menu[3].data,"Rom Image");
    strcpy(menu[4].data,"Main Menu");

    menu_cur = 0;
    while(sel!=-1)
    {
	tui_statusline("Memory Menu");
	Update(0);
	sel = tui_menubrowse(menu,5);

	switch(sel)
	{
	    case 0:
		fastmem_size=fastmem_size+2097152;
		if(fastmem_size>0x800000) { fastmem_size=0; }
		break;
	    case 1:
		chipmem_size=chipmem_size+0x80000;
		if(chipmem_size>0x200000) { chipmem_size=0; }
		break;
	    case 2:
		bogomem_size=bogomem_size+0x40000;
		if(bogomem_size>0x100000) { bogomem_size=0; }
		break;
	    case 3: strcpy(romfile, tuifilereq(".rom",romdefaultdir)); break;
	    case 4:
	    if((fastmem_size+chipmem_size+bogomem_size+1048576)>_go32_dpmi_remaining_physical_memory())
		tui_errorbox("Not enough free ram for these settings!");
	    else
		sel=-1;
	    break;
	}
    }
    menu_cur = 3;
}

void HDOptions(void)
{
    int i,vsel,count;
    long size=0;
    char mountvol[256], mountdir[256];

    int sel=0;
    char buff[256];

    struct bstring *menu,*vmenu;
    menu = (struct bstring *)calloc(7 , sizeof(struct bstring));

    strcpy(menu[0].data,"Hardfile Dis/Enable");
    strcpy(menu[1].data,"Add a virtual vol");
    strcpy(menu[2].data,"Add a virtual vol r/o");
    strcpy(menu[3].data,"Add a normal vol");
    strcpy(menu[4].data,"Add a normal vol r/o");
    strcpy(menu[5].data,"Remove a volume");
    strcpy(menu[6].data,"Main Menu");
    menu[1].val = 6;
    menu[2].val = 18;
    menu[3].val = 6;
    menu[4].val = 19;

    menu_cur = 0;
    while(sel!=-1)
    {
	tui_statusline("HardFile/volume menu");
	Update(0);
	sel = tui_menubrowse(menu,7);
	*mountvol = '\0';
	*mountdir = '\0';

	switch(sel)
	{
	case 0: currprefs.automount_uaedev=!currprefs.automount_uaedev; break;
	case 1:
	     if(tuiscan("Enter Volume name:",20,mountvol)) {
		 if(mountvol[strlen(mountvol)-1]==':')
		     mountvol[strlen(mountvol)-1] = 0;
		 if(tuiscan("Enter path of volume:",65,mountdir)) {
		     char mangle_buf[22];
		     int mangle_size = 8;

		     {
		     char *tmp;
		     while ((tmp = strchr(mountdir, '\\')))
			     *tmp = '/';
		     }
		     *mangle_buf = '\0';
		     if(tuiscan("Enter mangle size (default 8):",20,mangle_buf))
			 mangle_size=atoi(mangle_buf);
		     if((mangle_size<1) || (mangle_size>8))
			  mangle_size = 8;
		     add_filesys_unit(mountvol, mountdir, 0, 1, mangle_size);
		 }
	     }
	break;
	case 2:
	     if(tuiscan("Enter Volume name:",20,mountvol)) {
		 if(mountvol[strlen(mountvol)-1]==':')
		     mountvol[strlen(mountvol)-1] = 0;
		 if(tuiscan("Enter path of volume:",65,mountdir)) {
		     char mangle_buf[22];
		     int mangle_size = 8;

		     {
		     char *tmp;
		     while ((tmp = strchr(mountdir, '\\')))
			     *tmp = '/';
		     }
		     *mangle_buf = '\0';
		     if(tuiscan("Enter mangle size (default 8):",20,mangle_buf))
			 mangle_size=atoi(mangle_buf);
		     if((mangle_size<1) || (mangle_size>8))
			  mangle_size = 8;
		     add_filesys_unit(mountvol, mountdir, 1, 1, mangle_size);
		 }
	     }
	break;
	case 3:
	    if(tuiscan("Enter Volume name:",20,mountvol)) {
		if(mountvol[strlen(mountvol)-1]==':')
		    mountvol[strlen(mountvol)-1] = 0;
		if(tuiscan("Enter path of volume:",65,mountdir)) {
		    {
		    char *tmp;
		    while ((tmp = strchr(mountdir, '\\')))
			    *tmp = '/';
		    }
		    add_filesys_unit(mountvol, mountdir, 0, 0, 0);
		}
	    }
       break;
       case 4:
	    if(tuiscan("Enter Volume name:",20,mountvol)) {
		if(mountvol[strlen(mountvol)-1]==':')
		    mountvol[strlen(mountvol)-1] = 0;
		if(tuiscan("Enter path of volume:",65,mountdir)) {
		    {
		    char *tmp;
		    while ((tmp = strchr(mountdir, '\\')))
			    *tmp = '/';
		    }
		    add_filesys_unit(mountvol, mountdir, 1, 0, 0);
		}
	    }
       break;
       case 5:
	    if(tuiscan("Enter Volume name:",20,mountvol)) {
		if(mountvol[strlen(mountvol)-1]==':')
		    mountvol[strlen(mountvol)-1] = 0;
		if (kill_filesys_unit(mountvol) == -1)
		    tui_errorbox("Volume does not exist");
	    }
       break;
       case 6: sel=-1; break;
       }
    }
    menu_cur = 4;
}

void SoundOptions() {
    int sel=0;
    struct bstring *menu;
    char size[30];

    menu = (struct bstring *)calloc(8 , sizeof(struct bstring));

    strcpy(menu[0].data,"Produce Sound");
    strcpy(menu[1].data,"Bits 8/16");
    strcpy(menu[2].data,"Frequency");
    strcpy(menu[3].data,"Minimum Buffer Size");
    strcpy(menu[4].data,"Maximum Buffer Size");
    strcpy(menu[5].data,"Adjust Frequency");
    strcpy(menu[6].data,"Stereo Sound");
    strcpy(menu[7].data,"Main Menu");
    menu[3].val = 2;
    menu[4].val = 2;

    menu_cur = 0;
    while(sel!=-1)
    {
	tui_statusline("Sound Menu");
	Update(0);
	sel = tui_menubrowse(menu,8);
	*size = '\0';

	switch(sel)
	{
	    case 0:
		currprefs.produce_sound++;
		if(currprefs.produce_sound>3) { currprefs.produce_sound=0; }
		break;
	    case 1:
		if (currprefs.sound_bits == 8)
		    currprefs.sound_bits = 16;
		else
		    currprefs.sound_bits = 8;
		break;
	    case 2:
		if(tuiscan("Enter Frequency:",20,size))
		   currprefs.sound_freq = atoi(size);
		break;
	    case 3:
		if(tuiscan("Enter Minimum Size:",20,size))
		    currprefs.sound_minbsiz = atoi(size);
		break;
	    case 4:
		if(tuiscan("Enter Maximum Size:",20,size))
		    currprefs.sound_maxbsiz = atoi(size);
		break;
	    case 5:
		sound_adjust = !sound_adjust;
		break;
	    case 6:
		sound_stereo = !sound_stereo;
		break;
	    case 7: sel=-1; break;
	}
    }
    menu_cur = 5;
}

void OtherOptions() {
    int sel=0, joy1, joy2;

    struct bstring *menu;
    menu = (struct bstring *)calloc(8 , sizeof(struct bstring));

    menu[0].val=16; strcpy(menu[0].data,"Change Joystick 0");
    menu[1].val=16; strcpy(menu[1].data,"Change Joystick 1");
    strcpy(menu[2].data,"Parallel Port device");
    menu[3].val=12; strcpy(menu[3].data,"Serial Port device");
    strcpy(menu[4].data,"Serial on/off");
    strcpy(menu[5].data,"Emulation Accuracy");
    strcpy(menu[6].data,"68K Speed");
    strcpy(menu[7].data,"Main Menu");

    menu_cur = 0;
    while(sel!=-1)
    {
	tui_statusline("Other Options");
	Update(0);
	sel = tui_menubrowse(menu,8);

	switch(sel)
	{
	    case 0:
		joy1 = currprefs.fake_joystick & 255;
		joy2 = (currprefs.fake_joystick >> 8) & 255;
		joy1++;
		if(joy1>5) { joy1 = 0; }
		currprefs.fake_joystick = joy1 + (joy2 << 8);
	    break;
	    case 1:
		joy1 = currprefs.fake_joystick & 255;
		joy2 = (currprefs.fake_joystick >> 8) & 255;
		joy2++;
		if(joy2>5) { joy2 = 0; }
		currprefs.fake_joystick = joy1 + (joy2 << 8);
	    break;
	    case 2:
		tuiscan("Enter Parallel name:",20,prtname);
	    break;
	    case 3:
		tuiscan("Enter serial name:",20,sername);
	    break;
	    case 4:
		currprefs.use_serial=!currprefs.use_serial;
	    break;
	    case 5:
		currprefs.emul_accuracy++;
		if(currprefs.emul_accuracy>2) { currprefs.emul_accuracy = 0; }
	    break;
	    case 6:
		m68k_speed++;
		if(m68k_speed>16) { m68k_speed = 1; }
	    break;
	    case 7: sel=-1; break;
	}
    }
    menu_cur = 6;
}

void aboutuae(void)
{
    struct dialog dl;
    dl=dlog(dl,15,12,65,37,"About UAE!");
    gotoxy(2,1); cprintf("                    ");
    gotoxy(2,2); cprintf("                       ");
    gotoxy(2,3); cprintf("                   ");
    gotoxy(2,4); cprintf("                    ");
    gotoxy(2,5); cprintf("                   ");
    gotoxy(2,6); cprintf("            The Un*x Amiga Emulator!");
    gotoxy(2,7); cprintf("                 Version 0.6.9");
    gotoxy(2,9); cprintf("                 Dos Version by");
    gotoxy(2,10);cprintf("    Gustavo Goedert, Peter Remmers, Tim Gunn");
    gotoxy(2,11);cprintf("         Michael Sontheimer & Tomi Hakala");
    gotoxy(2,13);cprintf("    Documentation by Brett Eden & Nemo Pohle");
    gotoxy(2,16);cprintf("        UAE was written by Bernd Schmidt");
    gotoxy(2,17);cprintf("                      with");
    gotoxy(2,18);cprintf("Alessandro Bissacco, Hannu Rummukainen, Ed Hanway");
    gotoxy(2,19);cprintf("Denis Sablic, Marcus Sundberg, Samuel Devulder");
    gotoxy(2,20);cprintf("Stefan Reinauer, Marco Nelissen, Oliver Moehrke");
    gotoxy(2,21);cprintf("Olivier Raoul, Markus Gietzen, Christian Schmitt");
    gotoxy(2,22);cprintf("Herman ten Brugge, Tauno Taipaleenmaki.");
    gotoxy(2,25);cprintf("           Press any key to continue");
    getch();
    dlogdie(dl);
    menu_cur = 7;
}
