 /*
  * UAE - The Un*x Amiga Emulator
  *
  *  Serial Line Emulation
  *
  * (c) 1996, 1997 Stefan Reinauer <stepan@linux.de>
  * (c) 1997 Christian Schmitt <schmittc@uni-freiburg.de>
  *
  *  Thanks to Chris Hames <bitey@werple.net.au> for his hints.
  *
  */

#include "sysconfig.h"
#include "sysdeps.h"

#include "config.h"
#include "options.h"
#include "uae.h"
#include "memory.h"
#include "custom.h"
#include "readcpu.h"
#include "newcpu.h"
#include "cia.h"

#undef POSIX_SERIAL
/* Some more or less good way to determine whether we can safely compile in 
 * the serial stuff. I'm certain it breaks compilation on some systems. */
#if defined HAVE_SYS_TERMIOS_H && defined HAVE_POSIX_OPT_H && defined HAVE_SYS_IOCTL_H && defined HAVE_TCGETATTR
#define POSIX_SERIAL
#endif

#ifdef POSIX_SERIAL
#include <termios.h>
#include <unistd.h>
#include <sys/ioctl.h>
#endif

#if !defined B300 || !defined B230400
#undef POSIX_SERIAL
#endif

#ifndef O_NONBLOCK
#define O_NONBLOCK O_NDELAY
#endif

#define SERIALDEBUG 1 /* 0, 1, 2 3 */
#define MODEMTEST   1 /* 0 or 1 */

void serial_open(void);
void serial_close(void);
void serial_init(void);
void serial_exit(void);

void serial_dtr_on(void);
void serial_dtr_off(void);

uae_u16 serial_signals(int old, int new);

uae_u16 SERDATR(void);

void  SERDATS(void);
void  SERPER(uae_u16 w);
void  SERDAT(uae_u16 w);

int waitqueue=0,
    carrier=0,
    serdev=0,
    dsr=0,
    dtr=0,
    isbaeh=0;

#ifndef __DOS__
int sd = -1;
#else
FILE *sd = NULL;
#endif

#ifdef POSIX_SERIAL
    struct termios tios;
#endif

uae_u16 serper=0,serdat;

void SERPER(uae_u16 w)
{
    int baud=0,pspeed;

    if (!currprefs.use_serial)
	return;

#if defined POSIX_SERIAL
    if (serper == w)  /* don't set baudrate if it's already ok */
        return;
    serper=w;

    if (w&0x8000)
        fprintf (stdout, "SERPER: 9bit transmission not implemented.\n");

    switch (w & 0x7fff)
    {
     /* These values should be calculated by the current 
      * color clock value (NTSC/PAL). But this solution is
      * easy and it works.
      */

     case 0x2e9b:
     case 0x2e14: baud=300; pspeed=B300; break;
     case 0x170a:
     case 0x0b85: baud=1200; pspeed=B1200; break;
     case 0x05c2:
     case 0x05b9: baud=2400; pspeed=B2400; break;
     case 0x02e9:
     case 0x02e1: baud=4800; pspeed=B4800; break;
     case 0x0174:
     case 0x0170: baud=9600; pspeed=B9600; break;
     case 0x00b9:
     case 0x00b8: baud=19200; pspeed=B19200; break;
     case 0x005c:
     case 0x005d: baud=38400; pspeed=B38400; break;
     case 0x003d: baud=57600; pspeed=B57600; break;
     case 0x001e: baud=115200; pspeed=B115200; break;
     case 0x000f: baud=230400; pspeed=B230400; break;
     default:
      fprintf(stdout,"SERPER: unsupported baudrate (0x%04x) %d\n",w&0x7fff,
              (unsigned int)(3579546.471/(double)((w&0x7fff)+1)));  return;
    }

    /* Only access hardware when we own it */
    if (serdev == 1) {
      if(tcgetattr(sd,&tios)<0) {
        fprintf(stderr,"SERPER: TCGETATTR failed\n");
        return;
      }

      if(cfsetispeed(&tios,pspeed)<0) {    /* set serial input speed */
        fprintf(stderr,"SERPER: CFSETISPEED (%d bps) failed\n",baud);
        return;
      }
      if(cfsetospeed(&tios,pspeed)<0) {    /* set serial output speed */
        fprintf(stderr,"SERPER: CFSETOSPEED (%d bps) failed\n",baud);
        return;
      }

      if(tcsetattr(sd,TCSADRAIN,&tios)<0) {
        fprintf(stderr,"SERPER: TCSETATTR failed\n");
        return;
      }
    }
#endif

#if SERIALDEBUG > 0
  if (serdev == 1)
    fprintf(stdout,"SERPER: baudrate set to %d bit/sec\n",baud);
#endif
}

/* Not (fully) implemented yet:
 *  -  Something's wrong with the Interrupts. 
 *     (NComm works, TERM does not. TERM switches to a
 *     blind mode after a connect and wait's for the end
 *     of an asynchronous read before switching blind
 *     mode off again. It never gets there on UAE :-< )
 *
 *  -  RTS/CTS handshake, this is not really neccessary,
 *     because you can use RTS/CTS "outside" without
 *     passing it through to the emulated Amiga
 *
 *  -  ADCON-Register ($9e write, $10 read) Bit 11 (UARTBRK)
 *     (see "Amiga Intern", pg 246)
 *
 *  -  Serial interface still seems to eat some bytes... weird
 */

void SERDAT(uae_u16 w)
{
    unsigned char z;

    if (!currprefs.use_serial)
	return;

    z=(unsigned char)(w&0xff);

    if (currprefs.serial_demand && !dtr) {
      if (!isbaeh) {
        fprintf(stdout,"SERDAT: Baeh.. Your software needs SERIAL_ALWAYS to work properly.\n");
        isbaeh=1;
      }
      return;
    } else {
#ifndef __DOS__
      if (sd!=0) {
	write (sd,&z,1);
      }
#else
      if (sd!=NULL) {
	fputc (z, sd);
	fflush (sd);
      }
#endif
    }

#if SERIALDEBUG > 2
    fprintf(stderr,"SERDAT: wrote 0x%04x\n",w);
#endif

    serdat|=0x2000; /* Set TBE in the SERDATR ... */
    intreq|=1;      /* ... and in INTREQ register */
    return;
}

uae_u16 SERDATR(void)
{
    if (!currprefs.use_serial)
	return 0;
#if SERIALDEBUG > 2
    fprintf(stdout,"SERDATR: read 0x%04x\n",serdat);
#endif
    waitqueue = 0;
    return serdat;
}

void SERDATS(void)
{
    unsigned char z;
    int status;

    if ((!serdev)||(serdat&0x4000))
        return;

    if (waitqueue==1) {
      intreq|=0x0800;
      return;
    }

    if ((read (sd, &z,1))==1) {
      waitqueue=1;
      serdat=0x4100; /* RBF and STP set! */
      serdat|=((unsigned int)z)&0xff;
      intreq|=0x0800; /* Set RBF flag (Receive Buffer full) */

#if SERIALDEBUG > 1
      fprintf (stdout,"SERDATS: received 0x%02x --> serdat==0x%04x\n",
               (unsigned int)z, (unsigned int)serdat);
#endif
#ifdef POSIX_SERIAL
      ioctl(sd, TIOCMGET, &status);
      if (status & TIOCM_CAR) {
        if (!carrier) {
          CIA_shakehands_set(0x20); /* Push up Carrier Detect line */
          fprintf (stdout,"SERDATS: carrier detect.\n");
          carrier=1;
        }
      } else {
        if (carrier) {
          CIA_shakehands_clear(0x20); /* Carrier lost */
          fprintf (stdout,"SERDATS: carrier lost.\n");
          carrier=0;
        }
      }

      if (status & TIOCM_DSR) {
        if (!dsr) {
          CIA_shakehands_set(0x08); /* DSR on */
          dsr=1;
        }
      } else {
        if (dsr) {
          CIA_shakehands_clear(0x08); /* DSR off */
          dsr=0;
        }
      }
#endif
    }
}

uae_u16 serial_signals (int old, int new)
{
    if (((old&0x80)==0x80) && ((new&0x80)==0x00))
      serial_dtr_on();
    if (((old&0x80)==0x00) && ((new&0x80)==0x80))
      serial_dtr_off();

    if (((old&0x40)==0x40) && ((new&0x40)==0x00))
      fprintf (stdout,"RTS cleared.\n");
    if (((old&0x40)==0x00) && ((new&0x40)==0x40))
      fprintf (stdout,"RTS set.\n");
    if (((old&0x10)==0x10) && ((new&0x10)==0x00))
      fprintf (stdout,"CTS cleared.\n");
    if (((old&0x10)==0x00) && ((new&0x10)==0x10))
      fprintf (stdout,"CTS set.\n");

    return new; /* This value could also be changed here */
}

void serial_dtr_on(void)
{
  fprintf (stdout,"DTR on.\n");
  dtr=1;

  if (currprefs.serial_demand)
    serial_open();
}

void serial_dtr_off(void)
{
  fprintf (stdout,"DTR off.\n");
  dtr=0;
  if (currprefs.serial_demand)
    serial_close();
}

void serial_open(void)
{
    if (serdev == 1)
      return;

#ifndef __DOS__
    if ((sd=open(sername,O_RDWR|O_NONBLOCK))<0) {
      fprintf (stderr,"Error: Could not open Device %s\n",sername);
      return;
    }
#else
    if ((sd=fopen(sername,"wb"))==NULL) {
      fprintf (stdout,"Error: Could not open Device %s\n",sername);
      return;
    }
#endif

    serdev=1;

#ifdef POSIX_SERIAL
    if(tcgetattr(sd,&tios)<0) {		/* Initialize Serial tty */
      fprintf(stderr,"Serial: TCGETATTR failed\n");
      return;
    }

    cfmakeraw(&tios);

#ifndef MODEMTEST
    tios.c_cflag &= ~CRTSCTS; /* Disable RTS/CTS */
#else
    tios.c_cflag |= CRTSCTS; /* Enabled for testing modems */
#endif

    if(tcsetattr(sd,TCSADRAIN,&tios)<0) {
      fprintf(stderr,"Serial: TCSETATTR failed\n");
      return;
    }
#endif
}

void serial_close(void)
{
#ifndef __DOS__
    if (sd != 0)
      close(sd);
#else
    if (sd != NULL)
      fclose(sd);
#endif
    serdev=0;
}

void serial_init(void)
{
    if (!currprefs.use_serial)
	return;

    if (!currprefs.serial_demand)
        serial_open();

    serdat=0x2000;
    return;
}

void serial_exit(void)
{
    serial_close();	/* serial_close can always be called because it	*/
    dtr=0;		/* just closes *opened* filehandles which is ok	*/
    return;		/* when exiting.				*/
}
